-- Representation and display of regions and region automata

module RegionAutomaton (
   IntegerPartOrdering(..), FractionalPartOrdering(..), Region(..), State(..), RTransition(..), RAutomaton(..),
   isCClockConstraint, allOver, containsEqual, eqToBetween, bToEqual, bToEqual', 
   containsOtherEqual, clocksOver
) where

import Data.List
import TimedAutomaton

data IntegerPartOrdering = Equal (Clock,Int) | Between (Int,Clock,Int) | Over (Int,Clock)
data FractionalPartOrdering = Decreasing [[Clock]]
data Region = Region ([IntegerPartOrdering],FractionalPartOrdering)
data State = State (Location,Region)
data RTransition = RTransition { rsource::State, rlabel::Label, rtarget::State }
data RAutomaton = RAutomaton { rinitial::State, rgoals::[State], rstates::[State], rtransitions::[RTransition] }

instance Show RAutomaton where
   show a = "digraph RA {\n" ++
            (showStates (rinitial a) (rgoals a) (rstates a)) ++
            (concatMap (showRTransition) (rtransitions a)) ++ "}"

instance Eq IntegerPartOrdering where
   Equal (c,i) == Equal (c',i') = c == c' && i == i'
   Between (i,c,j) == Between (i',c',j') = i == i' && c == c' && j == j'
   Over (i,c) == Over (i',c') = i == i' && c == c'
   _ == _ = False

instance Ord IntegerPartOrdering where
   compare ipo1 ipo2 = compare (getClock ipo1) (getClock ipo2)

instance Show IntegerPartOrdering where
   show (Equal (c,i)) = (show c) ++ "=" ++ (show i)
   show (Between (i,c,i')) = (show i) ++ "<" ++ (show c) ++ "<" ++ (show i')
   show (Over (i,c)) = (show i) ++ "<" ++ (show c)

instance Eq FractionalPartOrdering where
   Decreasing l == Decreasing l' = l == l'

instance Show FractionalPartOrdering where
   show (Decreasing l) = show l

instance Eq Region where
   Region (i,f) == Region (i',f') = i == i' && f == f'

instance Show Region where
   show (Region (i,f)) = (show i) ++ " | " ++ (show f)

instance Eq State where
   State (Location l,Region r) == State (Location l',Region r') = l == l' && r == r' 

instance Show State where
   show (State (Location l,r)) = l ++ ", region: " ++ (show r) 

instance Eq RTransition where
   t == t' = (rsource t) == (rsource t') && (rlabel t) == (rlabel t') && (rtarget t) == (rtarget t')

showRTransition :: RTransition -> String
showRTransition t = (showState $ rsource t) ++ " -> " ++ 
                   (showState $ rtarget t) ++ " [label=\"" ++  
                   (showInfo [] (rlabel t) []) ++ "\"];\n"

showState :: State -> String
showState (State (Location l,r)) = l ++ (showRegion' r)

showStates :: State -> [State] -> [State] -> String
showStates i g a = (concatMap (showState') a') ++ (showRMarked i g)
   where
      a' = filter (\x -> not (x `elem` (i:g))) a

showState' :: State -> String
showState' s@(State (Location l,r)) = (showState s)  ++ " [label=\"" ++ l ++ "\\n" ++ (showRegion r) ++ "\"];\n"

showRegion :: Region -> String
showRegion (Region (l,Decreasing l')) = (concatMap (\x -> (showIPO x) ++ ", ") l) ++ "\\n" ++ (showFPO l')

showRegion' :: Region -> String
showRegion' (Region (l,Decreasing l')) = (concatMap (\x -> (showIPO' x)) l) ++ (showFPO' l')

showIPO :: IntegerPartOrdering -> String
showIPO (Equal (Clock c,i)) = c ++ " = " ++ (show i)
showIPO (Between (i,Clock c,i')) = (show i) ++ " < " ++ c ++ " < " ++ (show i')
showIPO (Over (i,Clock c)) = (show i) ++ " < " ++ c

showIPO' :: IntegerPartOrdering -> String
showIPO' (Equal (Clock c,i)) = c ++ "_" ++ (show i)
showIPO' (Between (i,Clock c,i')) = (show i) ++ "___" ++ c ++ "___" ++ (show i')
showIPO' (Over (i,Clock c)) = (show i) ++ "___" ++ c

showFPO :: [[Clock]] -> String
showFPO [] = ""
showFPO [l] = showEqual l
showFPO (l:ls) = (showEqual l) ++ " > " ++ (showFPO ls)

showFPO' :: [[Clock]] -> String
showFPO' [] = "empty"
showFPO' [l] = showEqual' l
showFPO' (l:ls) = (showEqual' l) ++ "__" ++ (showFPO' ls)

showEqual :: [Clock] -> String
showEqual [(Clock c)] = "f(" ++ c ++ ")"
showEqual ((Clock c):cs) = "f(" ++ c ++ ") = " ++ (showEqual cs)

showEqual' :: [Clock] -> String
showEqual' [(Clock c)] = c
showEqual' ((Clock c):cs) = c ++ "_" ++ (showEqual' cs)

showRMarked :: State -> [State] -> String
showRMarked s@(State (Location l, r)) [] = "dummy [style=invis,shape=none];\n" ++ 
                                               (showState s) ++ " [label=\"" ++ l ++ "\\n" ++ (showRegion r)  ++"\"];\n" ++ 
                                               "dummy -> " ++ (showState s) ++ "\n"
showRMarked s@(State (Location l, r)) ((s'@(State (Location l',r'))):locs) 
   | l' == l && r == r' =  "dummy [style=invis,shape=none];\n" ++ 
                           (showState s) ++ " [peripheries=2,label=\"" ++ l ++ "\\n" ++ (showRegion r) ++ "\"];\n" ++
                           "dummy -> " ++ (showState s) ++ ";\n" ++ 
                           (showRMarked' locs)
   | otherwise = (showState s') ++ " [peripheries=2,label=\"" ++ l' ++ "\\n" ++ (showRegion r') ++ "\"];\n" ++ 
                 (showRMarked s locs)
showRMarked' [] = "";
showRMarked' ((s@(State (Location l, r))):locs) = (showState s) ++ " [peripheries=2,label=\"" ++ 
                                                    l ++ "\\n" ++ (showRegion r) ++ "\"];\n" ++ 
                                                    (showRMarked' locs)

-- check if ipo is a constraint on c
isCClockConstraint :: Clock -> IntegerPartOrdering -> Bool
isCClockConstraint c (Equal (c',_)) = c' == c
isCClockConstraint c (Between (_,c',_)) = c' == c
isCClockConstraint c (Over (_,c')) = c' == c

-- check if ipos corresponds to a maximal region
allOver :: [IntegerPartOrdering] -> Bool
allOver [] = True
allOver ((Over (_,_)):ipos) = allOver ipos
allOver _ = False

-- check if ipos contains an equality constraint
containsEqual :: [IntegerPartOrdering] -> Bool
containsEqual [] = False
containsEqual ((Equal (_,_)):_) = True
containsEqual (ipo:ipos) = containsEqual ipos

-- change equal ipo into between or over ipo
eqToBetween :: Int -> IntegerPartOrdering -> IntegerPartOrdering
eqToBetween i (Equal (c,i')) 
   | i' < i = Between (i',c,i'+1)
   | otherwise = Over (i,c)
eqToBetween _ ipo = ipo

-- change between ipos with clocks in the set given as paramater into equal ipos
bToEqual :: [IntegerPartOrdering] -> [Clock] -> [IntegerPartOrdering]
bToEqual [] _ = []
bToEqual ipos [] = ipos
bToEqual ((ipo@(Between (i,c,i'))):ipos) cs
   | c `elem` cs = (Equal (c,i')) : (bToEqual ipos (filter (/= c) cs))
   | otherwise = ipo : (bToEqual ipos cs)
bToEqual (ipo:ipos) cs = ipo : (bToEqual ipos cs)

-- change all between ipos to equal ipos
bToEqual' :: [IntegerPartOrdering] -> [IntegerPartOrdering]
bToEqual' [] = []
bToEqual' ((ipo@(Between (i,c,i'))):ipos) = (Equal (c,i')) : (bToEqual' ipos)
bToEqual' (ipo:ipos) = ipo : (bToEqual' ipos)


-- check if ipos contains equal constraints on clocks out of the set given as parameter
containsOtherEqual :: [IntegerPartOrdering] -> [Clock] -> Bool
containsOtherEqual [] _ = False
containsOtherEqual ((Equal (c,_)):ipos) cs = (c `notElem` cs) || (containsOtherEqual ipos cs)
containsOtherEqual (ipo:ipos) cs = containsOtherEqual ipos cs

-- get the clock implied in an ipo
getClock :: IntegerPartOrdering -> Clock
getClock (Equal (c,_)) = c
getClock (Between (_,c,_)) = c
getClock (Over (_,c)) = c

-- get the ovec clocks in a list of ipos
clocksOver :: [IntegerPartOrdering] -> [Clock]
clocksOver ipos = clocksOver' ipos []
   where
      clocksOver' [] cs = cs
      clocksOver' ((Over (_,c)):ipos') cs = clocksOver' ipos' (c:cs)
      clocksOver' (_:ipos') cs = clocksOver' ipos' cs