-- Construction of region automaton

module RegionReach (
   regionAutomaton, strictRegionAutomaton
) where

import Data.List
import TimedAutomaton
import RegionAutomaton

-- Check if a constraint is satisfied in a region
isCompatibleWithRegion :: ClockConstraint -> Region -> Bool
isCompatibleWithRegion (EqualTo (c,i)) (Region (ipos,_)) = (Equal (c,i)) `elem` ipos
isCompatibleWithRegion (LowerThan (c,i)) (Region (ipos,_)) = 
   case find (isCClockConstraint c) ipos of
      Just (Equal (_,i')) -> i > i'
      Just (Between (_,_,i')) -> i >= i'
      Just (Over _) -> False
      Nothing -> False
isCompatibleWithRegion (GreaterThan (c,i)) (Region (ipos,_)) = 
   case find (isCClockConstraint c) ipos of
      Just (Equal (_,i')) -> i < i'
      Just (Between (i',_,_)) -> i <= i'
      Just (Over _) -> True
      Nothing -> False

-- Check if a transition is firable from a region
isFirableWithRegion :: Transition -> Region -> Bool
isFirableWithRegion t r = isFirableWithRegion' (guard t)
   where
      isFirableWithRegion' [] = True
      isFirableWithRegion' (g:gs) = (isCompatibleWithRegion g r) && (isFirableWithRegion' gs)

-- Compute the region obtained by reseting a set of clocks in a region
updateRegion :: Region -> [Clock] -> Region
updateRegion r [] = r
updateRegion (Region (ipos,Decreasing fpo)) l = Region (sort ipos',Decreasing fpo'')
   where
      (ipos',fpo') = updateRegion' l ipos fpo
      fpo'' = if fpo' == [] && (length l) <= 1 then [] else fpo'''
      fpo''' = if containsOtherEqual ipos' l then (init fpo') ++ (sort [(last fpo') ++ l]) else fpo' ++ (sort [l])
updateRegion' [] ipos fpo = (ipos,fpo)
updateRegion' (c:cs) ipos fpo = updateRegion' cs ipos' fpo'
   where
      ipos' = (Equal (c,0)):(filter (\x -> not (isCClockConstraint c x)) ipos)
      fpo' = filter (/=[]) (map (\x -> filter (/=c) x) fpo)

-- Compute the region obtained by firing a transition
fireTransitionFromRegion :: Transition -> Region -> Region
fireTransitionFromRegion t r = updateRegion r (resets t)

-- Compute the set of region reachable from a region by elapsing time only 
-- (the integer parameter gives the maximum bound in the automaton)
waitForRegion :: Region -> Int -> [Region]
waitForRegion r@(Region (ipos,Decreasing fpo)) i
   | allOver ipos = [r]
   | containsEqual ipos = r : (waitForRegion (Region (ipos',Decreasing fpo')) i)
   | otherwise = r : (waitForRegion (Region (sort ipos'',Decreasing fpo'')) i)
      where
         ipos' = sort (map (eqToBetween i) ipos)
         over = clocksOver ipos'
         removedOverClocks = filter (/=[]) (map (filter (`notElem` over)) fpo)
         fpo' = if (length removedOverClocks) == 1 && (length (head removedOverClocks)) == 1 then [] else removedOverClocks
         ipos'' = if fpo /= [] then bToEqual ipos (head fpo) else bToEqual' ipos
         fpo'' = if fpo /= [] then (tail fpo) ++ [(head fpo)] else []

-- Compute the set of region reachable from a region by elapsing time only (time must elapse) 
-- (the integer parameter gives the maximum bound in the automaton)
strictWaitForRegion :: Region -> Int -> [Region]
strictWaitForRegion r@(Region (ipos,Decreasing fpo)) i
   | containsEqual ipos = waitForRegion (Region (ipos',Decreasing fpo')) i
   | otherwise = (waitForRegion r i)
      where
         ipos' = sort (map (eqToBetween i) ipos)
         over = clocksOver ipos'
         removedOverClocks = filter (/=[]) (map (filter (`notElem` over)) fpo)
         fpo' = if (length removedOverClocks) == 1 && (length (head removedOverClocks)) == 1 then [] else removedOverClocks

-- Compute the RTransitions from a state and the states they reach
oneStepFun :: (Region -> Int -> [Region]) -> State -> Automaton -> Int -> ([RTransition],[State])
oneStepFun wForRegion s@(State (l,r)) a max = (rTransitions,states)
   where
      firable = filter (\t -> source t == l) (transitions a)
      startRegions = wForRegion r max
      rTransitions = oneStepFun' firable startRegions
      states = map (rtarget) rTransitions
      oneStepFun' [] _ = []
      oneStepFun' _ [] = []
      oneStepFun' (t:ts) rs = (oneStepFun'' t (filter (isFirableWithRegion t) rs)) ++ (oneStepFun' ts rs)
      oneStepFun'' t [] = []
      oneStepFun'' t (r':rs) = (RTransition { rsource=s, rlabel=(label t), rtarget=(State (target t,fireTransitionFromRegion t r'))}) : (oneStepFun'' t rs)

-- Compute the region automaton of a timed automaton
regionAutomatonFun :: (Region -> Int -> [Region]) -> Automaton -> RAutomaton
regionAutomatonFun wForRegion a = steps [init] [init] []
   where
      initIpos = map (\c -> Equal (c,0)) (clocks a)
      initIfo = Decreasing [(clocks a)]
      initRegion = Region (sort initIpos,initIfo)
      init = State (initial a, initRegion)
      max = maxConstant a
      steps [] statesSeen transitionsSeen = RAutomaton { 
                                               rinitial=init, 
                                               rgoals=(filter (\s@(State (l,_)) -> l `elem` (goals a)) statesSeen),
                                               rstates=statesSeen,
                                               rtransitions=transitionsSeen}
      steps (s:open) statesSeen transitionsSeen = steps open' statesSeen' transitionsSeen'
         where
            (newT,newS) = oneStepFun wForRegion s a max
            newS' = newS \\ statesSeen
            open' = open ++ newS'
            statesSeen' = newS' ++ statesSeen
            transitionsSeen' = union transitionsSeen newT 
            
            
regionAutomaton = regionAutomatonFun waitForRegion
strictRegionAutomaton = regionAutomatonFun strictWaitForRegion