-- Representation and display of zones and zone automata

module ZoneAutomaton (
   DBMVal(..), DBM(..), Zone(..), ZState(..), ZTransition(..), ZAutomaton(..),
   extend, intersectGuard, reset, isEmpty, include
) where

import Data.List
import TimedAutomaton
import Debug.Trace

data DBMVal = Infinity | Val Int | StrictVal Int
data DBM = DBM ([Clock],[[DBMVal]])
data Zone = Zone DBM
data ZState = ZState (Location,Zone)
data ZTransition = ZTransition { zsource::ZState, zlabel::Label, ztarget::ZState }
data ZAutomaton = ZAutomaton { zinitial::ZState, zgoals::[ZState], zstates::[ZState], ztransitions::[ZTransition] }

instance Show ZAutomaton where
   show a = "digraph RA {\n" ++ 
            showStates (zinitial a) (zgoals a) (zstates a) ++
            showTransitions (ztransitions a) ++ "}"

instance Eq ZState where
   ZState (l,z) == ZState (l',z') = l == l' && z == z'

instance Eq ZTransition where
   t == t' = (zsource t) == (zsource t') && (zlabel t) == (zlabel t') && (ztarget t) == (ztarget t')

instance Show Zone where
   show (Zone (DBM (cs,vs))) = (show cs) ++ ": " ++ (show vs)

instance Eq Zone where
   Zone d == Zone d' = d == d'

instance Eq DBM where
   DBM (_,vs) == DBM (_,vs') = vs == vs' 

instance Show DBMVal where
   show v = showVal v

instance Eq DBMVal where
   Infinity == Infinity = True
   Val i == Val i' = i == i'
   StrictVal i == StrictVal i' = i == i'
   _ == _ = False

showStates :: ZState -> [ZState] -> [ZState] -> String
showStates i gs [] = ""
showStates i gs (s:ss) 
   | s `elem` gs =
      if s == i then
         (showIFState s) ++ (showStates i (filter (/=s) gs) ss)
      else
         (showFState s) ++ (showStates i (filter (/=s) gs) ss)
   | otherwise =
      if s == i then
         (showIState s) ++ (showStates i gs ss)
      else
         (showState s) ++ (showStates i gs ss)

showTransitions :: [ZTransition] -> String
showTransitions [] = ""
showTransitions (t:ts) = (showState' $ zsource t) ++ " -> " ++
                         (showState' $ ztarget t) ++ " [ label=" ++
                         (showLabel $ zlabel t) ++ " ];\n" ++
                         (showTransitions ts)

showStateFun :: Bool -> Bool -> ZState -> String
showStateFun isInitial isFinal s@(ZState (Location l,Zone z)) = 
   (if isInitial then
      "dummy [ style=invis, shape=none ];\n"
   else
      ""
   ) ++   
   (showState' s) ++ " [ label=\"" ++
   l ++ "\\n" ++ (showDBM z) ++ "\"" ++
   (if isFinal then
      ", peripheries=2"
   else
      ""
   ) ++ " ];\n" ++
   (if isInitial then
      "dummy -> " ++ (showState' s) ++ ";\n"
   else
      ""
   ) 
 
showIFState = showStateFun True True
showIState = showStateFun True False
showFState = showStateFun False True
showState = showStateFun False False

showState' :: ZState -> String
showState'(ZState (Location l,Zone z)) = l ++ "_" ++ (showDBM' z)

showDBMFun :: (String -> [String] -> String) -> String -> (DBMVal -> String) -> DBM -> String
showDBMFun space newline showInt (DBM (cs,ls)) = (showLine (map getClockName cs) "") ++ (showLines cs ls)
   where
      showLines [] [] = ""
      showLines ((Clock c):cs) (l:ls) = (showLine (map showInt l) c) ++ (showLines cs ls)
      showLine is c = c ++ (space c (map getClockName cs)) ++ (showLine' is 0)
      showLine' [] _ = newline
      showLine' (i:is) k = i ++ (space i (map showInt (getColumn k ls))) ++ (showLine' is (k + 1))

getColumn :: Int -> [[a]] -> [a]
getColumn k ls = map (!!k) ls

showVal (Val i) = show i
showVal (StrictVal i) = (showVal (Val i)) ++ "s"
showVal Infinity = "inf"
spaceFun s ls = replicate k ' ' 
   where 
      k = (last (sort $ map length ls)) - (length s) + 3
showDBM = showDBMFun spaceFun "\\n" showVal

showVal' (Val i) = if i >= 0 then (show i) else "m" ++ (show (-i))
showVal' (StrictVal i) = (showVal' (Val i)) ++ "s"
showVal' Infinity = "infinity"
spaceFun' _ _ = "_"
showDBM' = showDBMFun spaceFun' "__" showVal'

showLabel :: Label -> String
showLabel (Label l) = l

-- Give the canonical form of a zone 
canonize :: Int -> Zone -> Zone
canonize maxConst (Zone (DBM (cs,vs))) = Zone (DBM (cs,vs'))
   where
      vs' = canonize' vs (length vs)
      canonize' m 0 = m
      canonize' m k = canonize' (matMult m vs) (k-1)
      matMult m m' = [[ foldl minVal Infinity $ zipWith addVal rowM colM' | colM' <- (transpose m') ] | rowM <- m ]
      addVal Infinity _ = Infinity
      addVal _ Infinity = Infinity
      addVal (Val i) (Val i') = 
         let i'' = i + i' in
            if i'' > maxConst then Infinity
            else if i'' < (- maxConst) then StrictVal (- maxConst)
            else Val i''
      addVal (Val i) (StrictVal i') =
         let i'' = i + i' in
            if i'' > maxConst then Infinity
            else if i'' < (- maxConst) then StrictVal (- maxConst)
            else StrictVal i''
      addVal (StrictVal i) (Val i') =
         let i'' = i + i' in
            if i'' > maxConst then Infinity
            else if i'' < (- maxConst) then StrictVal (- maxConst)
            else StrictVal i''
      addVal (StrictVal i) (StrictVal i') =
         let i'' = i + i' in
            if i'' > maxConst then Infinity
            else if i'' < (- maxConst) then StrictVal (- maxConst)
            else StrictVal i''

minVal :: DBMVal -> DBMVal -> DBMVal
minVal Infinity v = v
minVal v Infinity = v
minVal (Val i) (Val i') = if i < i' then Val i else Val i'
minVal (Val i) (StrictVal i') = if i < i' then Val i else StrictVal i'
minVal (StrictVal i) (Val i') = if i' < i then Val i' else StrictVal i
minVal (StrictVal i) (StrictVal i') = if i < i' then StrictVal i else StrictVal i'
      

-- Give the futur of a zone
extend :: Int -> Zone -> Zone
extend maxConst (Zone (DBM (cs,vs))) = canonize maxConst (Zone (DBM (cs,vs')))
   where
      vs'' = map (\l -> Infinity : (tail l)) vs
      fst = head $ head vs
      fst' = head vs''
      vs' = (fst : (tail fst')) : (tail vs'')

-- Intersect a zone with a set of guards
intersectGuard :: Int -> Zone -> [ClockConstraint] -> Zone
intersectGuard maxConst z [] = canonize maxConst z
intersectGuard maxConst z (cc:ccs) = intersectGuard maxConst z' ccs
   where
      z' = case cc of
              LowerThan (c,i) -> setC (StrictVal i) z [c]  
              GreaterThan (c,i) -> setL (StrictVal (-i)) z [c]
              EqualTo (c,i) -> (setC (Val i) (setL (Val (-i)) z [c]) [c]) 

-- Apply a reset of a set of clocks to a zone
setL :: DBMVal -> Zone -> [Clock] -> Zone
setL v z [] = z
setL v (Zone (DBM (cs,vs))) (cc:ccs) = setL v (Zone (DBM (cs,vs'))) ccs
   where
      vs' = case elemIndex cc cs of
               Just k -> 
                  let
                     (startHead,endHead) = splitAt k (head vs) 
                     newHead = startHead ++ (v : (tail endHead))
                  in
                     newHead : (tail vs)
               Nothing -> trace ("ZoneAutomaton.setL: not expected") []

setC :: DBMVal -> Zone -> [Clock] -> Zone
setC v z [] = z
setC v (Zone (DBM (cs,vs))) (cc:ccs) = setC v (Zone (DBM (cs,vs'))) ccs
   where
      vs' = case elemIndex cc cs of
               Just k -> 
                  let
                     (startAll,endAll) = splitAt k vs
                  in
                     startAll ++ ((v : (tail $ head endAll)) : (tail endAll))
               Nothing -> trace ("ZoneAutomaton.setC: not expected") []

propagateReset :: Zone -> [Clock] -> Zone
propagateReset z [] = z
propagateReset (Zone (DBM (cs,vs))) (cc:ccs) = propagateReset (Zone (DBM (cs,vs'))) ccs
   where
      vs' = case elemIndex cc cs of
         Just k ->
            let
               (startAll,endAll) = splitAt k vs
               vs'' = startAll ++ [(head vs)] ++ (tail endAll)
               vs''' = transpose vs''
               (startAll',endAll') = splitAt k vs'''
               vs'''' = startAll' ++ [(head vs''')] ++ (tail endAll')
            in
               transpose vs''''
         Nothing -> trace ("ZoneAutomaton.propagateReset: not expected") []


reset maxConst z cs = canonize maxConst (propagateReset (setC (Val 0) (setL (Val 0) z cs) cs) cs)

-- Checks emptyness of a zone (which should be canonical)
isEmpty :: Zone -> Bool
isEmpty (Zone (DBM (cs,vs))) = 
   filter (/=Val 0) (diagonal vs) /= []

-- Return the diagonal of a DBM
diagonal :: [[a]] -> [a]
diagonal dbm = diagonal' dbm [] 0
   where
      diagonal' [] d _ = reverse d
      diagonal' (l:ls) d k = diagonal' ls ((l!!k):d) (k+1)

-- Check if the first zone is included in the second
include :: Zone -> Zone -> Bool
include (Zone (DBM (cs,vs))) (Zone (DBM (cs',vs'))) = 
   not ((oneBelow (head vs) (head vs')) || (oneAbove (getColumn 0 vs) (getColumn 0 vs')))

-- Check if at least one of the clock values of the first line of
-- a dbm can be smaller than the same value of another dbm
oneBelow :: [DBMVal] -> [DBMVal] -> Bool
oneBelow [] [] = False
oneBelow (Infinity:vs) (_:vs') = oneBelow vs vs'
oneBelow ((Val i):vs) (v':vs') =
   case v' of
      Val i' -> i > i' || (oneBelow vs vs')
      StrictVal i' -> i >= i' || (oneBelow vs vs')
      Infinity -> True
oneBelow ((StrictVal i):vs) (v':vs') =
   case v' of
      Val i' -> i > i' || (oneBelow vs vs')
      StrictVal i' -> i > i' || (oneBelow vs vs')
      Infinity -> True

-- Check if a least one of the clock values of the first column of
-- a dbm can be greater than the same value of another dbm
-- The same as oneBelow because of the negative signs in the first line
oneAbove :: [DBMVal] -> [DBMVal] -> Bool
oneAbove = oneBelow